/*
 * Copyright 2020 Google LLC. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.empatnusabangsa.poseclassificationpoc.posedetector.classification;

import android.content.Context;
import android.media.AudioManager;
import android.media.ToneGenerator;
import android.os.Looper;
import android.util.Log;

import androidx.annotation.WorkerThread;

import com.empatnusabangsa.poseclassificationpoc.MainActivity;
import com.google.common.base.Preconditions;
import com.google.mlkit.vision.pose.Pose;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

/**
 * Accepts a stream of {@link Pose} for classification and Rep counting.
 */
public class PoseClassifierProcessor {
  private static final String TAG = "PoseClassifierProcessor";
  private static final String POSE_SAMPLES_FILE = "pose/poc_poses.csv";
//  private static final String POSE_SAMPLES_FILE = "pose/fitness_pose_samples.csv";

  // Specify classes for which we want rep counting.
  // These are the labels in the given {@code POSE_SAMPLES_FILE}. You can set your own class labels
  // for your pose samples.
  private static final String PUSHUPS_CLASS = "pushups_down";
  private static final String SQUATS_CLASS = "squats_down";
  private static final String SIT_CLASS = "sit";
  private static final String STAND_CLASS = "stand";
  private static final String FACE_LEFT = "face_left";
  private static final String FACE_RIGHT = "face_right";
  private static final String[] POSE_CLASSES = {
          STAND_CLASS, SIT_CLASS, FACE_LEFT, FACE_RIGHT
  };

  private static final String[] REPEAT_CLASSES = {
          SQUATS_CLASS
  };

  private final boolean isStreamMode;

  private com.empatnusabangsa.poseclassificationpoc.posedetector.classification.EMASmoothing emaSmoothing;
  private List<RepetitionCounter> repCounters;
  private PoseClassifier poseClassifier;
  private String lastRepResult;
  private String selectedPose;

  @WorkerThread
  public PoseClassifierProcessor(Context context, boolean isStreamMode, String selectedPose) {
    Preconditions.checkState(Looper.myLooper() != Looper.getMainLooper());
    this.isStreamMode = isStreamMode;
    this.selectedPose = selectedPose;
    if (isStreamMode) {
      emaSmoothing = new com.empatnusabangsa.poseclassificationpoc.posedetector.classification.EMASmoothing();
      repCounters = new ArrayList<>();
      lastRepResult = "";
    }
    loadPoseSamples(context);
  }

  private void loadPoseSamples(Context context) {
    List<com.empatnusabangsa.poseclassificationpoc.posedetector.classification.PoseSample> poseSamples = new ArrayList<>();
    try {
      BufferedReader reader = new BufferedReader(
          new InputStreamReader(context.getAssets().open(POSE_SAMPLES_FILE)));
      String csvLine = reader.readLine();
      while (csvLine != null) {
        // If line is not a valid {@link PoseSample}, we'll get null and skip adding to the list.
        com.empatnusabangsa.poseclassificationpoc.posedetector.classification.PoseSample poseSample = com.empatnusabangsa.poseclassificationpoc.posedetector.classification.PoseSample.getPoseSample(csvLine, ",");
        if (poseSample != null) {
          poseSamples.add(poseSample);
        }
        csvLine = reader.readLine();
      }
    } catch (IOException e) {
      Log.e(TAG, "Error when loading pose samples.\n" + e);
    }
    poseClassifier = new PoseClassifier(poseSamples);
    if (isStreamMode) {
      for (String className : REPEAT_CLASSES) {
        repCounters.add(new RepetitionCounter(className));
      }
    }
  }

  /**
   * Given a new {@link Pose} input, returns a list of formatted {@link String}s with Pose
   * classification results.
   *
   * <p>Currently it returns up to 2 strings as following:
   * 0: PoseClass : X reps
   * 1: PoseClass : [0.0-1.0] confidence
   */
  @WorkerThread
  public List<String> getPoseResult(Pose pose) {
    Preconditions.checkState(Looper.myLooper() != Looper.getMainLooper());
    List<String> result = new ArrayList<>();
    ClassificationResult classification = poseClassifier.classify(pose);

    // Update {@link RepetitionCounter}s if {@code isStreamMode}.
    if (isStreamMode) {
      // Feed pose to smoothing even if no pose found.
      classification = emaSmoothing.getSmoothedResult(classification);

      // Return early without updating repCounter if no pose found.
      if (pose.getAllPoseLandmarks().isEmpty()) {
        result.add(lastRepResult);
        return result;
      }

      Log.d(TAG, "getPoseResult: " + classification);
      Log.d(TAG, "pose: " + pose);





      if(selectedPose == SQUATS_CLASS) {
        for (RepetitionCounter repCounter : repCounters) {
          int repsBefore = repCounter.getNumRepeats();
          int repsAfter = repCounter.addClassificationResult(classification);
          if (repsAfter > repsBefore && repCounter.getClassName() == SQUATS_CLASS) {
            // Play a fun beep when rep counter updates.
            ToneGenerator tg = new ToneGenerator(AudioManager.STREAM_NOTIFICATION, 100);
            tg.startTone(ToneGenerator.TONE_PROP_BEEP);
            lastRepResult = String.format(
                    Locale.US, "%s : %d reps", repCounter.getClassName(), repsAfter);
            break;
          }
        }
        result.add(lastRepResult);
      } else {
        for(String classes : classification.getAllClasses()) {
          float classConfidence = classification.getClassConfidence(classes);
          if(classConfidence > 7.5f) {
            result.add(classes);
            result.add(String.valueOf(classification.getClassConfidence(classes)));
          }
        }
      }


    }



    // Add maxConfidence class of current frame to result if pose is found.
//    if (!pose.getAllPoseLandmarks().isEmpty()) {
//      String maxConfidenceClass = classification.getMaxConfidenceClass();
//      String maxConfidenceClassResult = String.format(
//          Locale.US,
//          "%s : %.2f confidence",
//          maxConfidenceClass,
//          classification.getClassConfidence(maxConfidenceClass)
//              / poseClassifier.confidenceRange());
//      result.add(maxConfidenceClassResult);
//    }

    return result;
  }

}
